import math
import warnings
from typing import Sequence

from rlgym_tools.rocket_league.reward_functions.auto_reward_normalizer.running_normalizer import RunningNormalizer


class SimpleZNormalizer(RunningNormalizer):
    def __init__(self, alpha=0.999999, clip=20, center=True, scale=True, ignore_zeros=False):
        self.alpha = alpha
        self.clip = clip
        self.center = center
        self.scale = scale
        self.ignore_zeros = ignore_zeros

        if self.ignore_zeros and self.center:
            warnings.warn(
                "If ignoring zeros, it is recommended that you disable centering "
                "to avoid excessive negative values and keep the sign of the unnormalized reward.",
                UserWarning,
            )

        self._count = 0
        self._sum = 0
        self._sq_sum = 0

    def update(self, values: Sequence):
        # Update the statistics. We do this first to avoid division by zero
        if self.ignore_zeros:
            values = [v for v in values if v != 0]
        c = len(values)
        if c == 0:
            return
        s = sum(values)
        ss = sum(r ** 2 for r in values)

        self._count = self.alpha * self._count + (1 - self.alpha) * c
        self._sum = self.alpha * self._sum + (1 - self.alpha) * s
        self._sq_sum = self.alpha * self._sq_sum + (1 - self.alpha) * ss

    def normalize(self, values: Sequence):
        # Calculate the mean and variance
        if self._count == 0:
            return [math.nan] * len(values)  # The user has to deal with these
        mean = self._sum / self._count if self.center else 0
        var = (self._sq_sum / self._count) - (mean ** 2) if self.scale else 1

        # Normalize the rewards
        std = math.sqrt(var) + 1e-8

        norm_rewards = []
        for reward in values:
            z = (reward - mean) / std
            clipped = min(max(z, -self.clip), self.clip)
            norm_rewards.append(clipped)

        return norm_rewards



from typing import List, Dict, Any

from rlgym.api import RewardFunction, AgentID, StateType
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.reward_functions.auto_reward_normalizer.running_normalizer import RunningNormalizer
from rlgym_tools.rocket_league.reward_functions.auto_reward_normalizer.simple_z_normalizer import SimpleZNormalizer


class AutoRewardNormalizer(RewardFunction[AgentID, GameState, float]):

    # Update modes:
    ON_RESET = 0
    BEFORE_NORM = 2
    AFTER_NORM = 3
    BEFORE_NORM_THEN_ON_RESET = 4

    def __init__(self, reward_function: RewardFunction[AgentID, GameState, float],
                 dist_tracker: RunningNormalizer = None,
                 update_mode: int=None):
        self.reward_function = reward_function
        self.dist_tracker = dist_tracker or SimpleZNormalizer()
        if update_mode is None:
            if isinstance(self.dist_tracker, SimpleZNormalizer):
                self.update_mode = self.BEFORE_NORM_THEN_ON_RESET
            else:
                self.update_mode = self.ON_RESET
        self.update_mode = update_mode

        self._rewards_since_reset = []

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        self.reward_function.reset(agents, initial_state, shared_info)
        if len(self._rewards_since_reset) > 0:
            if self.update_mode == self.ON_RESET:
                self.dist_tracker.update(self._rewards_since_reset)
            elif self.update_mode == self.BEFORE_NORM_THEN_ON_RESET:
                self.update_mode = self.ON_RESET
        self._rewards_since_reset = []

    def get_rewards(self, agents: List[AgentID], state: StateType, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = self.reward_function.get_rewards(agents, state, is_terminated, is_truncated, shared_info)

        rew_list = [rewards[a] for a in agents]

        match self.update_mode:
            case self.ON_RESET:
                self._rewards_since_reset += rew_list
                norm_rewards = self.dist_tracker.normalize(rew_list)
            case self.BEFORE_NORM:
                self.dist_tracker.update(rew_list)
                norm_rewards = self.dist_tracker.normalize(rew_list)
            case self.AFTER_NORM:
                norm_rewards = self.dist_tracker.normalize(rew_list)
                self.dist_tracker.update(rew_list)
            case self.BEFORE_NORM_THEN_ON_RESET:
                self._rewards_since_reset += rew_list
                self.dist_tracker.update(rew_list)
                norm_rewards = self.dist_tracker.normalize(rew_list)
            case _:
                raise ValueError(f"Unknown update mode: {self.update_mode}")

        norm_rewards = {
            a: norm_rewards[i]
            for i, a in enumerate(agents)
        }

        if any(abs(r) == 10 for r in norm_rewards.values()):
            debug = True

        return norm_rewards



from dataclasses import dataclass

import RocketSim as rsim
import numpy as np
from rlgym.rocket_league.api import Car


@dataclass(frozen=True, slots=True)
class Hitbox:
    length: float
    width: float
    height: float

    offset_x: float
    offset_y: float
    offset_z: float

    front_wheel_radius: float
    front_suspension_rest: float
    front_wheel_offset_x: float
    front_wheel_offset_y: float
    front_wheel_offset_z: float

    back_wheel_radius: float
    back_suspension_rest: float
    back_wheel_offset_x: float
    back_wheel_offset_y: float
    back_wheel_offset_z: float

    @property
    def corner_offsets(self):
        # Center of mass to each corner
        # Order will be:
        # back left bottom, back left top, back right bottom, back right top,
        # front left bottom, front left top, front right bottom, front right top
        mid_length = self.length / 2
        mid_width = self.width / 2
        mid_height = self.height / 2
        naive = np.array([
            [a * mid_length, b * mid_width, c * mid_height]
            for a in (-1, 1)
            for b in (-1, 1)
            for c in (-1, 1)
        ])
        offset = np.array([self.offset_x, self.offset_y, self.offset_z])
        result = naive + offset
        result.setflags(write=False)
        return result

    def corner_positions(self, car: Car):
        physics = car.physics
        position = physics.position
        rot_mtx = physics.rotation_mtx

        return (self.corner_offsets @ rot_mtx.T) + position


OCTANE = Hitbox(
    length=120.507,
    width=86.6994,
    height=38.6591,
    offset_x=13.87566,
    offset_y=0,
    offset_z=20.755,
    front_wheel_radius=12.50,
    front_suspension_rest=38.755,
    front_wheel_offset_x=51.25,
    front_wheel_offset_y=25.90,
    front_wheel_offset_z=20.755,
    back_wheel_radius=15.00,
    back_suspension_rest=37.055,
    back_wheel_offset_x=-33.75,
    back_wheel_offset_y=29.50,
    back_wheel_offset_z=20.755
)

DOMINUS = Hitbox(
    length=130.427,
    width=85.7799,
    height=33.8,
    offset_x=9,
    offset_y=0,
    offset_z=15.75,
    front_wheel_radius=12.00,
    front_suspension_rest=33.95,
    front_wheel_offset_x=50.30,
    front_wheel_offset_y=31.10,
    front_wheel_offset_z=15.75,
    back_wheel_radius=13.50,
    back_suspension_rest=33.85,
    back_wheel_offset_x=-34.75,
    back_wheel_offset_y=33.00,
    back_wheel_offset_z=15.75
)

PLANK = Hitbox(
    length=131.32,
    width=87.1704,
    height=31.8944,
    offset_x=9.00857,
    offset_y=0,
    offset_z=12.0942,
    front_wheel_radius=12.50,
    front_suspension_rest=31.9242,
    front_wheel_offset_x=49.97,
    front_wheel_offset_y=27.80,
    front_wheel_offset_z=12.0942,
    back_wheel_radius=17.00,
    back_suspension_rest=27.9242,
    back_wheel_offset_x=-35.43,
    back_wheel_offset_y=20.28,
    back_wheel_offset_z=12.0942
)

BREAKOUT = Hitbox(
    length=133.992,
    width=83.021,
    height=32.8,
    offset_x=12.5,
    offset_y=0,
    offset_z=11.75,
    front_wheel_radius=13.50,
    front_suspension_rest=29.7,
    front_wheel_offset_x=51.50,
    front_wheel_offset_y=26.67,
    front_wheel_offset_z=11.75,
    back_wheel_radius=15.00,
    back_suspension_rest=29.666,
    back_wheel_offset_x=-35.75,
    back_wheel_offset_y=35.00,
    back_wheel_offset_z=11.75
)

HYBRID = Hitbox(
    length=129.519,
    width=84.6879,
    height=36.6591,
    offset_x=13.8757,
    offset_y=0,
    offset_z=20.755,
    front_wheel_radius=12.50,
    front_suspension_rest=38.755,
    front_wheel_offset_x=51.25,
    front_wheel_offset_y=25.90,
    front_wheel_offset_z=20.755,
    back_wheel_radius=15.00,
    back_suspension_rest=37.055,
    back_wheel_offset_x=-34.00,
    back_wheel_offset_y=29.50,
    back_wheel_offset_z=20.755
)

MERC = Hitbox(
    length=123.22,
    width=79.2103,
    height=44.1591,
    offset_x=11.3757,
    offset_y=0,
    offset_z=21.505,
    front_wheel_radius=15.00,
    front_suspension_rest=39.505,
    front_wheel_offset_x=51.25,
    front_wheel_offset_y=25.90,
    front_wheel_offset_z=21.505,
    back_wheel_radius=15.00,
    back_suspension_rest=39.105,
    back_wheel_offset_x=-33.75,
    back_wheel_offset_y=29.50,
    back_wheel_offset_z=21.505
)

HITBOXES = {
    rsim.CarConfig.OCTANE: OCTANE,
    rsim.CarConfig.DOMINUS: DOMINUS,
    rsim.CarConfig.PLANK: PLANK,
    rsim.CarConfig.BREAKOUT: BREAKOUT,
    rsim.CarConfig.HYBRID: HYBRID,
    rsim.CarConfig.MERC: MERC,
}



import sqlite3
from typing import Sequence

from rlgym_tools.rocket_league.reward_functions.auto_reward_normalizer.simple_z_normalizer import SimpleZNormalizer


class SQLiteZNormalizer(SimpleZNormalizer):
    # Shares stats between processes via SQLite.
    def __init__(self, db_path: str, name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.db_path = db_path
        self.name = name
        self.conn = sqlite3.connect(db_path, isolation_level=None)
        self.cursor = self.conn.cursor()
        with self.conn:
            self.cursor.execute(f"CREATE TABLE IF NOT EXISTS stats ("
                                f"name TEXT PRIMARY KEY, "
                                f"count REAL, "
                                f"sum REAL, "
                                f"sq_sum REAL"
                                f")")
            self.cursor.execute(f"INSERT OR IGNORE INTO stats (name, count, sum, sq_sum) VALUES (?, ?, ?, ?)",
                                (self.name, 0, 0, 0))

            self.cursor.execute(f"SELECT count, sum, sq_sum FROM stats WHERE name=?", (self.name,))
            row = self.cursor.fetchone()
            self._count, self._sum, self._sq_sum = row

    def __del__(self):
        self.conn.close()

    def update(self, values: Sequence):
        # Get latest stats from the database in case they were updated by another process
        with self.conn:
            row = self.conn.execute(f"SELECT count, sum, sq_sum FROM stats WHERE name=?", (self.name,)).fetchone()

            self._count, self._sum, self._sq_sum = row

            super().update(values)

            self.conn.execute("""
                UPDATE stats SET count = ?, sum = ?, sq_sum = ? WHERE name = ?
            """, (self._count, self._sum, self._sq_sum, self.name))

    # normalize() is inherited from SimpleZNormalizer



import math
from enum import IntEnum, auto

import numpy as np
from rlgym.rocket_league.common_values import CEILING_Z, SIDE_WALL_X, BACK_WALL_Y

CORNER_BOUNDING_BOX = 8064  # corners are defined by |x|+|y|=8064


class Surface(IntEnum):
    GROUND = auto()
    CEILING = auto()
    SIDE_WALL_NEG_X = auto()
    SIDE_WALL_POS_X = auto()
    BACK_WALL_NEG_Y = auto()
    BACK_WALL_POS_Y = auto()
    CORNER_NEG_X_NEG_Y = auto()
    CORNER_POS_X_NEG_Y = auto()
    CORNER_NEG_X_POS_Y = auto()
    CORNER_POS_X_POS_Y = auto()

    # OTHER = auto()

    @property
    def is_back_wall(self):
        return self in (self.BACK_WALL_NEG_Y, self.BACK_WALL_POS_Y)

    @property
    def is_side_wall(self):
        return self in (self.SIDE_WALL_NEG_X, self.SIDE_WALL_POS_X)

    def naive_dist_from(self, pos):
        # Calculates "naive" distance to a position.
        # E.g. is uses mathematical definitions of the field bounds
        # It will give accurate distance whenever the current surface is the closest
        x, y, z = pos
        match self:
            # Ground and ceiling are defined by z=0 and z=2044
            case Surface.GROUND:
                dist = z
            case Surface.CEILING:
                dist = CEILING_Z - z

            # Side walls are defined by |x|=5120, so the distance is simply 5120±x
            case Surface.SIDE_WALL_NEG_X:
                dist = SIDE_WALL_X + x
            case Surface.SIDE_WALL_POS_X:
                dist = SIDE_WALL_X - x

            # Back walls are defined by |y|=5120, so the distance is simply 5120±y
            case Surface.BACK_WALL_NEG_Y:
                dist = BACK_WALL_Y + y
            case Surface.BACK_WALL_POS_Y:
                dist = BACK_WALL_Y - y

            # The lines defining the corners are Ax+By+C=0
            # Where A=±1, B=±1, C=8064
            # The distance to the line is given by the formula
            # |Ax+By+C|/sqrt(A^2+B^2)
            # Which simplifies to
            # |±x±y+8064|/sqrt(2)
            # with different signs for the four quadrants
            case Surface.CORNER_NEG_X_NEG_Y:
                dist = abs(x + y + CORNER_BOUNDING_BOX) / math.sqrt(2)
            case Surface.CORNER_POS_X_NEG_Y:
                dist = abs(-x + y + CORNER_BOUNDING_BOX) / math.sqrt(2)
            case Surface.CORNER_NEG_X_POS_Y:
                dist = abs(x - y + CORNER_BOUNDING_BOX) / math.sqrt(2)
            case Surface.CORNER_POS_X_POS_Y:
                dist = abs(-x - y + CORNER_BOUNDING_BOX) / math.sqrt(2)

            # Default to nan
            case _:
                dist = math.nan

        # Abs just to make sure, should only really apply when inside the goal
        return abs(dist)


def closest_surface(pos: np.ndarray) -> (Surface, float):
    closest = min(Surface, key=lambda s: s.naive_dist_from(pos))
    dist = closest.naive_dist_from(pos)
    return closest, dist



# From https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/version.py
# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '2.3.5'


release_notes = {
    '2.3.5': """
    - Fix bad pitch/yaw/roll error calculation in pick_action.py
    """,
    '2.3.4': """
    - Improve conditions in pick_action.py
    """,
    '2.3.3': """
    - Fix changing dictionary entries in ReplayFrame actions after yield.
    """,
    '2.3.2': """
    - In pick_action.py: better logic for when jump matters, and fix for throttle logic when no boost.
    """,
    '2.3.1': """
    - Fix flip reset obtain detection in FlipResetReward
    """,
    '2.3.0': """
    - Update RLGym requirement to >=2.0.1
    - Add AutoRewardNormalizer,SimpleZNormalizer and SQLiteNormalizer for automatically normalizing rewards
    - Add hitboxes.py containing a Hitbox dataclass and the standard in-game hitboxes
    - Add surface.py with a Surface enum methods for getting distance to different kinds of arena surfaces
    - Fix AugmentMutator not adjusting flip torque and autoflip drection
    - Fix DemoReward rewarding the same demo multiple times
    """,
    '2.2.5': """
    - Add new carball.exe
    """,
    '2.2.4': """
    - Fix dodge torque adjustment by normalizing torque in replay converter.
    """,
    '2.2.3': """
    - Fix pitch and yaw on dodges by normalizing torque in replay converter.
    """,
    '2.2.2': """
    - Fix RandomPhysicsMutator giving invalid rotation and disabling ball collision.
    """,
    '2.2.1': """
    - Fix error in RandomPhysicsMutator
    """,
    '2.2.0': """
    - Add V1 objects and V2 to V1 conversion functions.
    """,
    '2.1.2': """
    - Fix activation function in BoostKeepReward
    - Add checks to RandomPhysicsMutator to make sure we place cars and ball inside the field.
    - Use copying in AugmentMutator to circumvent a bug in base RLGym library when using KickoffMutator
    - Fix DelayedAction for delays longer than action length
    - Set delayed_actions in shared_info on reset in DelayedAction
    - Set go_to_kickoff and is_over in RandomScoreboardMutator
    """,
    '2.1.1': """
    - Improve accuracy of the replay to RLGym converter
    """,
    '2.1.0': """
    - Add reward reporting tools for inspecting rewards with replays
    - Add some convenience methods to the Action dataclass
    """,
    '2.0.3': """
        - Update MANIFEST.in to include carball.exe
    """,
    '2.0.2': """
       - Add carball.exe to package_data
    """,
    '2.0.1': """
       - Fix serialization (added wheel contacts and padded replay frame support, and tested with new script)
       - Fix ReplayMutator
       - Add RandomPhysicsMutator
    """,
    '2.0.0': """
       Move to RLGym v2:
       - Remove all v1 code
       - Add action parsers: ActionHistoryWrapper, AdvancedLookupTableAction, DelayedAction, QueuedAction
       - Add done conditions: BallHitGroundCondition, CarOnGroundCondition, GameCondition
       - Add math utils for ball, gamma, inverse aerial controls, relative physics, skellam distribution and solid angle
       - Add Action dataclass and serialization
       - Add obs builder: RelativeDefaultObs
       - Add renderer: RocketSimVisRenderer
       - Add new replay parser and replay converter to new ReplayFrame object
       - Add rewards: AdvancedTouchReward, AerialDistanceReward, BallTravelReward, BoostChangeReward, BoostKeepReward, 
                      DemoReward, EpisodeEndReward, FlipResetReward, GoalProbReward, StackReward, 
                      TeamSpiritRewardWrapper, VelocityPlayerToBallReward, WaveDashReward
       - Add shared info providers: BallPredictionProvider, MultiProvider, ScoreboardProvider, SerializedProvider
       - Add state mutators: AugmentMutator, ConfigMutator, GameMutator, HitboxMutator, RandomScoreboardMutator, 
                             ReplayMutator, VariableTeamSizeMutator, WeightedSampleMutator
       - Add transition engine: MultiEnvEngine
    """,
    '1.8.3': """
       - Update parse_actions in lookup_act.py (Jeff)
    """,
    '1.8.2': """
       - Fix no touch timer in GameCondition (Rolv)
       - Update RLLib example (Aech)
    """,
    '1.8.1': """
        - Refactor GameCondition (Rolv, Impossibum)
        - Fix a small mistake in LookupAction (Rolv)
    """,
    '1.8.0': """
        - Add lookup parser as used by Nexto/Tecko (Rolv)
        - Add customizable odds to WallPracticeState (Soren)
        - Add code for reducing SB3 model size for RLBot botpack, with example (DI3D)
        - Update AdvancedPadder for RLGym 1.2 (Kaiyotech)
        - Update example code for RLGym 1.2 (mstuettgen)
        - Fix AdvancedStacker (Some Rando)
        - Fix broken imports in SequentialRewards (Some Rando)
        - Fix bad indent in JumpTouchReward (Some Rando)
        """,
    '1.7.0':
        """
        - Add AdvancedObsPadder (Impossibum)
        - Add JumpTouchReward (Impossibum)
        - Fix NameError in KickoffReward (benjamincburns)
        - Add generate_probabilities as a method to ReplaySetter (Rolv)
        - Upgrade WallSetter (Soren) and fix bug when num_cars == 1 (Kaiyotech)
        - Add max overtime to GameCondition (Rolv)
        """,
    '1.6.6':
        """
        -WallStateSetter now has airdribble setups and harder wall plays
        """,
    '1.6.5':
        """
        -GoalieStateSetter now better emulates incoming airdribbles
        -fixed WallStateSetter bug and increased starting diversity
        """,
    '1.6.4':
        """
        -Added wall play state setter (Soren)
        """,
    '1.6.3':
        """
        - Fix hoops-like setter for multiple players (Carrot)
        """,
    '1.6.2':
        """
        - Added hoops-like setter (Carrot)
        - Fixed kickoff-like setter (Carrot)
        - Fixed pitch in KBMAction (Rolv)
        - Added include_frame option in replay converter (wrongu)
        """,
    '1.6.1':
        """
        - Fixed angular velocities in replay augmenter
        """,
    '1.6.0':
        """
        - Added GoaliePracticeState setter (Soren)
        - Added ReplaySetter (Carrot)
        - Added AugmentSetter (NeverCast)
        - Fixed an error in replay converter
        """,
    '1.5.3':
        """
        - Yet another fix for GameCondition
        """,
    '1.5.2':
        """
        - Another fix for GameCondition
        """,
    '1.5.1':
        """
        - Fix for GameCondition
        """,
    '1.5.0':
        """
        - Add SB3CombinedLogReward (LuminousLamp)
        - Add SB3InstantaneousFPSCallback (LuminousLamp)
        - Add KickoffLikeSetter (Carrot)
        - Add GameCondition (Rolv)
        """,
    '1.4.1':
        """
        - Remove SB3MultiDiscreteWrapper
        - Update SB3 examples to include action parsers
        """,
    '1.4.0':
        """
        - Add KBM action parser
        """,
    '1.3.0':
        """
        - Add KickoffReward (Impossibum)
        - Add SequentialReward (Impossibum)
        - Better individual reward logging for SB3 (LuminousLamp)
        - Add launch_preference as parameter to SB3MultipleInstanceEnv
        """,
    '1.2.0':
        """
        - Added multi-model support for SB3
        - Added weighted sample setter
        - Added general stacker
        - Various bug fixes
        """,
    '1.1.4':
        """
        - Bug fixes for AdvancedStacker
        """,
    '1.1.3':
        """
        - Fix RLGym 1.0 incompatibility
        """,
    '1.1.2':
        """
        - Fix version import (was absolute, now relative)
        """,
    '1.1.1':
        """
        - Fixed an invalid Python 3.7 import
        - AnnealRewards now has initial_count parameter
        """,
    '1.1.0':
        """
        - Added functionality in SB3MultipleInstanceEnv to take multiple match objects (one for each instance).
        - Added AnnealRewards for transitioning between reward functions.
        - Added granularity option to SB3MultiDiscreteWrapper.
        - Added negative slope parameter to DiffReward (negative values are multiplied by this).
        """,
    '1.0.0':
        """
        - Added replay to rlgym GameState converter
        - Moved SB3 environments from rlgym (now called SB3SingleInstanceEnv and SB3MultipleInstanceEnv) and fixed some bugs
        - Added SB3MultiDiscreteWrapper, SB3DistributeRewardsWrapper and SB3LogReward 
        - Added extra reward functions (DiffReward, DistributeRewards and MultiplyRewards)
        - Added RLLibEnv
        - Added working example code for SB3 and RLlib
        """
}


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def print_current_release_notes():
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print("")
